local classes = require("lua/explosionsNG/helpers/classes")

--- @alias ExplosiveState "ready" | "active" | "finished"
--- @alias AnyExplosive Explosive<any, any>

--- __Abstract__
---
--- The superclass that all explosive subclasses will inherit from.
---
--- The constructor takes in a parameter called `presetMapper`. This is used when the configuration specifies a
--- preset. Subclasses calling this method should implement the `getPresetOptions` and pass this method as this
--- parameter (the actual function can be a wrapper, but should make use of/call this method). In normal OOP, this would
--- be called within this constructor, however, as Lua does not support OOP (only mimicry), this method when called as
--- `self:getPresetOptions` will refer to this class as the parameter `self`, not the subclass. Therefore, it cannot be
--- used. See `getPresetOptions` for more details on that method.
---
--- __Generics__
--- - `T` is the type for the preset of the explosive
--- - `S` is the type of the options of the explosive
--- @class Explosive<T, S>
--- @field private state ExplosiveState The current state of the explosive
--- @field protected options S The options for the explosive
--- @field protected position LuaVec3 The in-game position of the explosive
--- @field protected onFinish (fun(): void) | nil A function to call when this explosive's state is set to `"Finished"`
local Explosive = classes.newClass()

--- Optional method that can be overridden in subclasses.
--- Called when the explosive is detonated.
--- @protected
function Explosive:onDetonation() end

--- Optional method that can be overridden in subclasses.
--- Called when the explosive has finished exploding.
--- @protected
function Explosive:onExplosionFinished() end

--- __Abstract__
---
--- Whether the detonation of this explosive has now finished.
--- Note that this should be determined independently of the `state` of this explosive
--- @return boolean
--- @protected
function Explosive:isDetonationFinished()
  error("isDetonationFinished is not implemented")
  return false
end

--- __Abstract__
--- @param dt number The time since the last frame
--- @protected
function Explosive:updateDetonation(dt)
  error("updateDetonation is not implemented")
end

--- __Abstract__
---
--- Returns a given explosive options object with values for the given `preset`.
---
--- An implementation of this method __must__ remain static in nature (i.e. not access any instance fields or methods)
--- in order for this method to be called correctly in the constructor. This method is defined as an instance method
--- purely for the generic types to apply.
---
--- This means it is therefore safe to statically access this method.
--- @param preset T The preset of the explosive to get the options of
--- @return S The options of the explosive
function Explosive:getPresetOptions(preset)
  error("getPresetOptions is not implemented")
  return nil
end

--- Sets `state` to `"active"`.
--- Calls `onDetonation`.
function Explosive:detonate()
  self.state = "active"
  self:onDetonation()
end

--- Sets `state` to `"finished"`.
--- Calls `onExplosionFinished`.
--- If the `onFinish` callback is defined, this method calls it.
--- @protected
function Explosive:postExplosion()
  self.state = "finished"
  self:onExplosionFinished()
  if self.onFinish then self.onFinish() end
end

--- If the call to `isActive` returns false, this method does nothing.
--- Otherwise, this method calls `updateDetonation` and `updateDetonation`.
--- If the call to `isDetonationFinished` returns true, this method calls `postExplosion`.
--- @param dt number The time since the last frame
function Explosive:update(dt)
  if not self:isActive() then return end

  self:updateDetonation(dt)

  if self:isDetonationFinished() then self:postExplosion() end
end

--- @return boolean Whether the `state` is `"active"`
--- @private
function Explosive:isActive()
  return self.state == "active"
end

--- Sets `onFinish` to the given `callback` function. Whenever this explosive's `state` changes to `"Finished"`, this
--- callback function will be called
--- @param callback fun(): void The callback function
function Explosive:registerOnFinishedHandler(callback)
  -- Currently, multiple callbacks are not needed and are treated as an error
  if self.onFinish then error("This explosive already has an onFinish callback") end
  self.onFinish = callback
end

--- __Constructor__
---
--- __Generics__
--- - `U` is the type for the preset of the explosive
--- - `V` is the type of the options of the explosive
--- @generic U, V
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: U | nil, options: V | nil } The configuration of the explosive. Either properties can be given; if both are given then the `preset` is ignored
--- @param presetMapper fun(self: any, preset: U): V A mapping function that produces an options object from a preset. `self` is unused
--- @return Explosive<U, V> The constructed instance
--- @protected
function Explosive:new(position, configuration, presetMapper)
  local self = --[[--- @type Explosive<U, V> ]] setmetatable({}, Explosive)
  self.options = configuration.options and configuration.options or presetMapper(self, configuration.preset)
  self.position = vec3(position)
  self.state = "ready"
  self.onFinish = nil

  if not self.options then
    log("E", "explosivesNG", "No options have been provided with the following configuration: "..dumps(configuration))
  end

  return self
end

return Explosive
