local classes = require("lua/explosionsNG/helpers/classes")
local Explosive = require("lua/explosionsNG/explosives/base/Explosive")
local Lightning = require("lua/explosionsNG/worldObjects/environment/Lightning")
local ShockwaveExplosive = require("lua/explosionsNG/explosives/shockwave/ShockwaveExplosive")

--- Properties:
---  - `shockwave` - The list of options for the shockwave part of the explosive
---  - `lightning` - The list of options for the lightning part of the explosive
--- @alias LightningExplosiveOptions { shockwave: ShockwaveExplosiveOptions | nil, lightning: LightningOptions | nil }
--- @alias LightningExplosivePreset "default"

--- @class LightningExplosive : Explosive<LightningExplosivePreset, LightningExplosiveOptions>
--- @field private lightning Lightning The lightning bolt
--- @field private shockwave ShockwaveExplosive The shockwave explosion
local LightningExplosive = classes.newClass(Explosive)

--- @param preset LightningExplosivePreset The preset of the explosive to get the options of
--- @return LightningExplosiveOptions The options for the `ShockwaveExplosive` as part of a lightning explosive
--- @see Explosive#getPresetOptions
function LightningExplosive:getPresetOptions(preset)
  return {
    shockwave = {
      time = 1, maxRadius = 15, forceMultiplier = 3
    }
  }
end

--- @see Explosive#onDetonation
function LightningExplosive:onDetonation()
  self.shockwave:detonate()
end

--- @return boolean Whether the explosive has finished
--- @see Explosive#isDetonationFinished
function LightningExplosive:isDetonationFinished()
  return self.lightning:isFinished() and self.shockwave:isDetonationFinished()
end

--- @param dt number The time since the last frame
--- @see Explosive#updateDetonation
function LightningExplosive:updateDetonation(dt)
  self.lightning:update(dt)
  self.shockwave:update(dt)
end

--- __Constructor__
--- @param position LuaVec3 The in-game position of the explosive
--- @param configuration { preset: LightningExplosivePreset | nil, options: LightningExplosiveOptions | nil } The configuration of the explosive. This is either a preset or list of options.
--- @return LightningExplosive The constructed instance
function LightningExplosive:new(position, configuration)
  local self = classes.setmetatable(Explosive:new(position, configuration, LightningExplosive.getPresetOptions), LightningExplosive)
  self.lightning = Lightning:new(position, self.options.lightning)
  self.shockwave = ShockwaveExplosive:new(position, { options = self.options.shockwave })
  return self
end

return LightningExplosive
