local Array = require("lua/explosionsNG/misc/Array")

--- @alias ExplosiveIdentifier { uiId: string, explosiveId: string, presetId: string }
--- @alias ExplosiveWithIdentifier { identifier: ExplosiveIdentifier, explosive: AnyExplosive }
--- @alias BossData { progress: ConvergencePhaseProgress[], failed: boolean }

local explosives = Array:new(--[[--- @type ExplosiveWithIdentifier[] ]] {})

--- Removes the given `explosive` from the `explosives` Array
--- @param explosive AnyExplosive The explosive to remove
local function removeExplosive(explosive)
  local foundExplosive = explosives:findBy(function(explosiveWithId) return explosiveWithId.explosive == explosive end)

  if not foundExplosive then
    log("W", "explosivesNG", "Could not remove explosive as it doesn't exist")
    return
  end

  explosives:remove(foundExplosive)
  guihooks.trigger("engExplosiveFinished", foundExplosive.identifier.uiId)
end

--- Registers a handler to the finish event of the explosive and detonates the explosive.
--- This function will also adds the explosive to this handlers internal array.
--- @param uiId string The UI id to store with the explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @param explosive AnyExplosive The explosive to add
local function addExplosive(uiId, explosiveId, presetId, explosive)
  local identifier = { uiId = uiId, explosiveId = explosiveId, presetId = presetId }
  explosives:add({ identifier = identifier, explosive = explosive })
  explosive:registerOnFinishedHandler(function() removeExplosive(explosive) end)
  explosive:detonate()
end

--- Updates each explosive in `explosives`
--- @param dt number The time since the last frame
local function update(dt)
  explosives:forEach(function(explosiveWithId) explosiveWithId.explosive:update(dt) end)
end

--- @return Array<ExplosiveIdentifier> The list of all ids for each explosive being handled by this handler
local function getIds()
  return explosives:map(function(explosiveWithId) return explosiveWithId.identifier end)
end

--- @return BossData The current progress through the "The Convergence" boss battle. `nil` if no boss is currently active.
local function getBossData()
  local convergence = explosives:findBy(function(explosiveWithId) return explosiveWithId.identifier.explosiveId == "convergence" end)
  if not convergence then return nil end

  local castedConvergence = --[[--- @type ConvergenceExplosive ]] convergence.explosive
  return { progress = castedConvergence:getCurrentProgress(), failed = castedConvergence:hasFailedCurrentPhase() }
end

--- Clears all explosives in the `explosives` array
local function clear()
  explosives:clear()
end

return {
  removeExplosive = removeExplosive,
  addExplosive = addExplosive,
  update = update,
  getIds = getIds,
  getBossData = getBossData,
  clear = clear
}