local vehicles = require("lua/explosionsNG/handlers/vehicles")
local explosivesFacade = require("lua/explosionsNG/stores/explosives/facade")
local Array = require("lua/explosionsNG/misc/Array")
local Preview = require("lua/explosionsNG/worldObjects/Preview")

--- @alias CoordinateType "player" | "camera"

local previews = Array:new(--[[--- @type Preview[] ]] {})

--- Returns the preview object which matches the given `id`
--- @param id string The id of the preview to get
--- @return Preview The preview with matching id. `nil` if no preview exists
local function findPreview(id)
  return previews:findBy(function(preview) return preview:getId() == id end)
end

--- Returns the position of either the player's current vehicle, or the camera position, based on the given
--- `coordinateType`.
--- @param coordinateType CoordinateType What object to use the coordinates of
--- @return LuaVec3 The position of the player or camera
local function getPositionOfObject(coordinateType)
  if coordinateType == "player" then
    local playerVehicle = vehicles.getPlayerVehicle()
    if playerVehicle then return playerVehicle:getPosition() end
    log("W", "explosivesNG", "Player vehicle does not exist, defaulting to the camera position.")
  end

  return core_camera.getPosition()
end

--- Returns the offset that should be added to the elevation of an object with the given bounds
--- @param bounds BoundingArea | nil The bounds
--- @return LuaVec3 The elevation offset of the bounds
local function getElevationOffset(bounds)
  if not bounds then return vec3() end
  if type(bounds) == "number" then return vec3(0, 0, --[[--- @type number ]] bounds) end

  local lowestBound = Array:new(--[[--- @type SphericalBound[] ]] bounds):findByComparison(function(a, b)
    return a.offset.z - a.radius < b.offset.z - b.radius and a or b
  end)
  -- The lowest z value will be the offset value plus the radius of the sphere itself to get the bottommost point on the sphere
  return vec3(0, 0, lowestBound.offset.z + lowestBound.radius)
end

--- Sets the preview with corresponding `id`'s position with the given `position`
--- @param id string The id of the preview to update the coordinates of
--- @param position Vector The position to set the preview to
local function updatePreviewPosition(id, position)
  local preview = findPreview(id)

  if not preview then
    log("E", "explosivesNG", "Could not update preview as the preview with id="..tostring(id).." does not exist")
    return
  end

  preview:updatePosition(vec3(position))
end

--- Sets the preview with corresponding `id`'s position and returns the position that was used. Note that this function
--- respects the preview's bounds (if it has one).
--- @param id string The id of the preview to update the coordinates of
--- @param type CoordinateType What object to use the coordinates of
--- @return LuaVec3 The position of the player or camera that was used to update the preview
local function updatePreviewPositionByType(id, type)
  local basePosition = getPositionOfObject(type)

  local preview = findPreview(id)
  local boundingRange = preview:getBoundsRange()
  local elevation = getElevationOffset(boundingRange)

  -- Move the preview upwards if it has a bounding box
  local position = basePosition + elevation
  preview:updatePosition(position)
  return position
end

--- Creates a new preview and sets the position to the given `position`. This function will return the numerical id of
--- the newly created preview.
--- @param position Vector The position to set the preview to
--- @param explosiveId string The explosive id
--- @param presetId string The preset's id of the explosive
--- @return string The id of the created preview
local function createPreviewByPosition(position, explosiveId, presetId)
  local stats = explosivesFacade.getAbsoluteExplosivePresetStats(explosiveId, presetId)
  local boundingRadius = explosivesFacade.getBoundingArea(explosiveId, presetId)

  local allExplosives = explosivesFacade.getListOfExplosives()
  local thisExplosive = Array:new(allExplosives):findBy(function(explosive) return explosive.id == explosiveId end)

  local newPreview = Preview:new(vec3(position), thisExplosive.tier, stats.range, boundingRadius)
  previews:add(newPreview)
  return newPreview:getId()
end

--- Creates a new preview and sets the position to the given `CoordinateType`'s object position. This function will
--- return the numerical id of the newly created preview.
--- @param explosiveId string The explosive id
--- @param presetId string The preset's id of the explosive
--- @param type CoordinateType What object to use the coordinates of
--- @return { id: string, position: Vector } The id and position of the created preview
local function createPreviewByOption(type, explosiveId, presetId)
  local basePosition = getPositionOfObject(type)

  local boundingRange = explosivesFacade.getBoundingArea(explosiveId, presetId)
  local elevation = getElevationOffset(boundingRange)

  -- Move the explosive upwards if it has a bounding box
  local position = basePosition + elevation
  local previewId = createPreviewByPosition(position, explosiveId, presetId)

  return { id = previewId, position = position:toDict() }
end

--- Deletes the preview that matches the given `id`
--- @param id string The id of the preview to delete
local function deletePreview(id)
  local preview = findPreview(id)

  if not preview then
    log("E", "explosivesNG", "Could not delete preview as the preview with id="..tostring(id).." does not exist")
    return
  end

  previews:remove(preview)
  preview:delete()
end

--- Clears all previews
local function clear()
  previews:clear()
end

return {
  updatePreviewPosition = updatePreviewPosition,
  updatePreviewPositionByType = updatePreviewPositionByType,
  createPreviewByPosition = createPreviewByPosition,
  createPreviewByOption = createPreviewByOption,
  deletePreview = deletePreview,
  clear = clear
}