local Vehicle = require("lua/explosionsNG/vehicles/Vehicle")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias PlayerVehicleSpawnedFn fun(player: Vehicle): void

local vehicles = Array:new(--[[--- @type Vehicle[] ]] {})
local playerCallbacks = Array:new(--[[--- @type PlayerVehicleSpawnedFn[] ]] {})
--- A flag of whether the player is currently not in a vehicle
local isPlayerNotInAVehicle = true

--- Returns a predicate that matches the given vehicle `id`
--- @param id number The vehicle id that the predicate should match
--- @return fun(vehicle: Vehicle): boolean
local function vehicleIdPredicate(id)
  return function(vehicle) return vehicle:getId() == id end
end

--- Creates a new `Vehicle` object using the game object found using the given `id`.
--- If the given `id` does not yet have  a `Vehicle` associated with it, this function adds it to the internal array.
--- Otherwise, this function will overwrite the existing object.
--- @param id number The id of the vehicle to add
local function addVehicle(id)
  local predicate = vehicleIdPredicate(id)
  local newVehicle = Vehicle:new(id, --[[--- @type VehicleSceneObject ]] be:getObjectByID(id))
  local existingVehicle = vehicles:findBy(predicate)

  if existingVehicle then
    vehicles:replaceBy(newVehicle, predicate)
    log("I", "explosivesNG", "Replaced existing Vehicle instance for "..tostring(id))
    return
  end

  vehicles:add(newVehicle)
end

--- Checks and returns whether at least one vehicle is within the given range of the given position
--- @param position LuaVec3 The centre of the range to check
--- @param range number The extent of the range (as a radius)
--- @return boolean whether at least one vehicle is within the given range of the given position
local function isVehicleInRange(position, range)
  return vehicles:some(function(vehicle) return vehicle:isWithin(position, range) end)
end

--- @return Array<Vehicle> A copy of the array of all vehicles
local function getVehicles()
  return vehicles:clone()
end

--- Returns the vehicle with the given id
--- @param id number The id of the vehicle to find
--- @return Vehicle The found vehicle. `nil` if no vehicle exists
local function getVehicle(id)
  return vehicles:findBy(vehicleIdPredicate(id))
end

--- Finds and returns the vehicle the player is in
--- @return Vehicle The found player vehicle. `nil` if the player is not in any vehicle
local function getPlayerVehicle()
  return vehicles:findBy(function(vehicle) return vehicle:isPlayer() end)
end

--- It should be noted that this function will check specifically whether there is a vehicle that this handler knows
--- about that reports it is a player vehicle; it is independent of the `isPlayerNotInAVehicle` flag.
--- @return boolean Whether the player is in a vehicle or not
local function isPlayerInAVehicle()
  return getPlayerVehicle() ~= nil
end

--- Stores the given callback function to this handler. When a player vehicle is spawned, this callback will be run.
---
--- __Important:__ it should be noted that once the callback is triggered, it will be cleared from this handler. If a
--- caller wishes to maintain a link to this event, after the callback is triggered they should re-register the
--- callback.
--- @param callback PlayerVehicleSpawnedFn The handler callback
local function registerOnPlayerSpawnedHandler(callback)
  playerCallbacks:add(callback)
end

--- Checks whether a player vehicle exists. If it does, this method will then call all callbacks that have been
--- registered to this handler before clearing them and setting the `isPlayerNotInAVehicle` flag to `false.
local function checkPlayerNowSpawned()
  local playerVehicle = getPlayerVehicle()

  if playerVehicle then
    playerCallbacks:forEach(function(callback) callback(playerVehicle) end)
    playerCallbacks:clear()

    isPlayerNotInAVehicle = false
  end
end

--- Updates every vehicle in this handler
--- @param dt number The time since the last frame
local function update(dt)
  vehicles:forEach(function(vehicle) vehicle:update(dt) end)
end

--- A vehicle has been removed.
--- Removes the given vehicle from the vehicles array, given the vehicle's id.
--- @param id number The id of the vehicle to remove
local function onVehicleDestroyed(id)
  vehicles = vehicles:removeIf(vehicleIdPredicate(id))
  isPlayerNotInAVehicle = not isPlayerInAVehicle()
end

--- The player vehicle has been changed.
--- Notifies the UI as to whether the player is currently in a vehicle or not.
local function onVehicleSwitched()
  if isPlayerNotInAVehicle then
    checkPlayerNowSpawned()
  end

  guihooks.trigger("engIsPlayerInAVehicle", isPlayerInAVehicle())
end

--- A vehicle has been reset.
--- Resets the vehicle with the given in-game object id
--- @param id number The id of the vehicle
local function onVehicleResetted(id)
  local vehicle = getVehicle(id)

  if not vehicle then
    log("W", "explosivesNG", "A vehicle has been reset that does not exist on this handler.")
    return
  end

  vehicle:reset()
end

--- A vehicle has been spawned.
--- Creates a new vehicle to this handler with the in-game object that matches the given id.
--- @param id number The id of the vehicle
local function onVehicleSpawned(id)
  addVehicle(id)
end

--- Removes any vehicles that no longer exist in-game.
--- This should be called if a vehicle is removed and the id is not known, i.e. loading a new level.
local function clear()
  vehicles = vehicles:removeIf(Vehicle.isRemovedFromLevel)
end

return {
  isVehicleInRange = isVehicleInRange,
  getVehicles = getVehicles,
  getPlayerVehicle = getPlayerVehicle,
  isPlayerInAVehicle = isPlayerInAVehicle,
  registerOnPlayerSpawnedHandler = registerOnPlayerSpawnedHandler,
  update = update,
  onVehicleDestroyed = onVehicleDestroyed,
  onVehicleSwitched = onVehicleSwitched,
  onVehicleResetted = onVehicleResetted,
  onVehicleSpawned = onVehicleSpawned,
  clear = clear
}
