local collisions = require("lua/explosionsNG/helpers/collisions")
local CircularPath = require("lua/explosionsNG/pathing/CircularPath")

--- Returns a random offset position from the origin. The algorithm used by this function will create a circle of random
--- radius and return a position on that circle at a random angle. The fuzziness parameter will increase the radius of
--- this theoretical circle, resulting in a high spread of offset positions from the origin.
--- @param fuzziness number Controls how
--- @return LuaVec3 A random offset position
local function getRandomOffset(fuzziness)
  local radius = math.random() * fuzziness
  local angle = math.random() * math.pi * 2
  local circular = CircularPath:new(vec3(), radius, angle)
  return circular:getCurrentPosition():z0() -- Don't add any randomness to the altitude
end

--- Predicts the position a vehicle will be at in the future. This function is purposefully basic and simply uses the
--- linear interpolation of the vehicles current direction and velocity.
---
--- The fuzziness parameter will decrease the accuracy of the returned position, and will return a random deviation
--- around the predicted point, upto and including this value in metres.
--- @param vehicle Vehicle The vehicle to predict a new position for
--- @param time number The time at which to predict the position of
--- @param fuzziness number Controls how accurate the returned position is. Higher values in less accuracy.
--- @return LuaVec3 The predicted position of the vehicle
local function predictPosition(vehicle, time, fuzziness)
  local currentPosition = vehicle:getPosition()
  local linearPrediction = vehicle:getVelocity() * time

  local offset = getRandomOffset(fuzziness)
  -- Find a suitable XY position
  local targetPositionXY = currentPosition + linearPrediction + offset
  -- Find a suitable altitude
  local altitude = collisions.findElevationOfCollision(targetPositionXY.x, targetPositionXY.y)

  return vec3(targetPositionXY.x, targetPositionXY.y, altitude)
end

return {
  predictPosition = predictPosition
}