local ShockwaveExplosive = require("lua/explosionsNG/explosives/shockwave/ShockwaveExplosive")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/shockwave/ShockwaveExplosiveStats")
local AnomalyExplosive = require("lua/explosionsNG/explosives/anomaly/AnomalyExplosive")
local AnomalyExplosiveStats = require("lua/explosionsNG/explosives/anomaly/AnomalyExplosiveStats")
local ClusterExplosive = require("lua/explosionsNG/explosives/cluster/ClusterExplosive")
local ClusterExplosiveStats = require("lua/explosionsNG/explosives/cluster/ClusterExplosiveStats")
local LightningStormExplosive = require("lua/explosionsNG/explosives/lightningStorm/LightningStormExplosive")
local LightningStormExplosiveStats = require("lua/explosionsNG/explosives/lightningStorm/LightningStormExplosiveStats")
local LandMineExplosive = require("lua/explosionsNG/explosives/landMine/LandMineExplosive")
local LandMineExplosiveStats = require("lua/explosionsNG/explosives/landMine/LandMineExplosiveStats")
local SupernovaExplosive = require("lua/explosionsNG/explosives/supernova/SupernovaExplosive")
local SupernovaExplosiveStats = require("lua/explosionsNG/explosives/supernova/SupernovaExplosiveStats")
local ConvergenceExplosive = require("lua/explosionsNG/explosives/convergence/ConvergenceExplosive")
local ConvergenceExplosiveStats = require("lua/explosionsNG/explosives/convergence/ConvergenceExplosiveStats")
local GravityOrbsExplosive = require("lua/explosionsNG/explosives/gravityOrbs/GravityOrbsExplosive")
local GravityOrbsExplosiveStats = require("lua/explosionsNG/explosives/gravityOrbs/GravityOrbsExplosiveStats")
local OrbitalStrikeExplosive = require("lua/explosionsNG/explosives/orbitalStrike/OrbitalStrikeExplosive")
local OrbitalStrikeExplosiveStats = require("lua/explosionsNG/explosives/orbitalStrike/OrbitalStrikeExplosiveStats")

-- Intelligent type checking with all these generics has not been implemented as luanalysis is nowhere near intelligent enough for it :-)

local EXPLOSIVE_DETAILS = --[[--- @type table<string, { explosive: any, stats: any }> ]] {
  shockwave = {
    explosive = ShockwaveExplosive,
    stats = ShockwaveExplosiveStats
  },
  gravityOrbs = {
    explosive = GravityOrbsExplosive,
    stats = GravityOrbsExplosiveStats
  },
  cluster = {
    explosive = ClusterExplosive,
    stats = ClusterExplosiveStats
  },
  lightningStorm = {
    explosive = LightningStormExplosive,
    stats = LightningStormExplosiveStats
  },
  orbitalStrike = {
    explosive = OrbitalStrikeExplosive,
    stats = OrbitalStrikeExplosiveStats
  },
  landMine = {
    explosive = LandMineExplosive,
    stats = LandMineExplosiveStats
  },
  anomaly = {
    explosive = AnomalyExplosive,
    stats = AnomalyExplosiveStats
  },
  supernova = {
    explosive = SupernovaExplosive,
    stats = SupernovaExplosiveStats
  },
  convergence = {
    explosive = ConvergenceExplosive,
    stats = ConvergenceExplosiveStats
  }
}

--- Returns the stats instance for a given preset of an explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id (preset) of the preset of the explosive
--- @return ExplosiveStats<any> The stats of the preset for the explosive
local function getExplosiveStatsInstance(explosiveId, presetId)
  local detail = EXPLOSIVE_DETAILS[explosiveId]
  local options = detail.explosive:getPresetOptions(presetId)
  return --[[--- @type ExplosiveStats<any> ]] detail.stats:new(options)
end

--- Returns the stats for a given preset of an explosive
--- @param explosiveId string The id of the explosive
--- @param presetId string The id (preset) of the preset of the explosive
--- @return PresetStats The stats of the preset for the explosive
local function getAbsoluteExplosivePresetStats(explosiveId, presetId)
  local stats = getExplosiveStatsInstance(explosiveId, presetId)
  return stats:getAbsoluteStats()
end

--- Returns the stats for a given preset of an explosive, scaled to values between [0, 1]
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @return PresetStats The stats of the preset for the explosive
local function getExplosivePresetStats(explosiveId, presetId)
  local stats = getExplosiveStatsInstance(explosiveId, presetId)
  return stats:getNormalisedStats()
end

--- Creates and returns the explosive given its corresponding ids and the position to place it at
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @return AnyExplosive The created explosive
local function createNewExplosive(explosiveId, presetId, position)
  local detail = EXPLOSIVE_DETAILS[explosiveId]
  return --[[--- @type AnyExplosive ]] detail.explosive:new(position, { preset = presetId })
end

--- Returns the bounding area of objects that will be created for a given explosive and preset. This function will
--- return `nil` for explosives that do not create long-lasting objects as part of their explosions.
--- @param explosiveId string The id of the explosive
--- @param presetId string The id of the preset of the explosive
--- @return BoundingArea | nil The bounding area range of the explosive. `nil` if there is no bounding area for the explosive.
local function getBoundingArea(explosiveId, presetId)
  -- TODO Abusive typecasting
  if explosiveId == "anomaly"       then return     AnomalyExplosive:getBoundingArea(--[[--- @type AnomalyExplosivePreset     ]] presetId) end
  if explosiveId == "supernova"     then return   SupernovaExplosive:getBoundingArea(--[[--- @type SupernovaExplosivePreset   ]] presetId) end
  if explosiveId == "convergence"   then return ConvergenceExplosive:getBoundingArea(--[[--- @type ConvergenceExplosivePreset ]] presetId) end
  return nil
end

return {
  getAbsoluteExplosivePresetStats = getAbsoluteExplosivePresetStats,
  createNewExplosive = createNewExplosive,
  getBoundingArea = getBoundingArea,
  getExplosivePresetStats = getExplosivePresetStats
}