local collisions = require("lua/explosionsNG/helpers/collisions")
local vehiclesHandler = require("lua/explosionsNG/handlers/vehicles")
local explosivesHandler = require("lua/explosionsNG/handlers/explosives")
local worldObjectsHandler = require("lua/explosionsNG/handlers/worldObjects/handler")
local previewsHandler = require("lua/explosionsNG/handlers/previews")
local explosivesFacade = require("lua/explosionsNG/stores/explosives/facade")
local Array = require("lua/explosionsNG/misc/Array")

--- @alias RequestParams { explosiveId: string, presetId: string, uiId: string, position: Vector }

--- A flag that stores whether the UI needs resetting.
--- For example: a new freeroam level was loaded and the UI configuration should not persist.
local uiNeedsResetting = false

--- Wraps the given hook in a function that will not call the hook on calling, should the player not be in freeroam
--- @param hook function The hook to wrap
--- @return function The wrapped hook
local freeroamHookWrapper = function(hook)
  return function(...)
    if not extensions.freeroam_freeroam.state.freeroamActive then return end
    hook(...)
  end
end

--- @param dt number The time since the last frame
local function onPreRender(dt)
  if simTimeAuthority.getPause() then return end

  dt = dt * bullettime.get()

  explosivesHandler.update(dt)
  vehiclesHandler.update(dt)
end

--- Creates and places the explosives that matches the given list of explosive and preset ids as well as the position to
--- place each explosive.
--- @param requests RequestParams[] The request parameters containing the details for each explosive
local function placeExplosives(requests)
  Array:new(requests):forEach(function(request)
    --request.position = vec3(-485, 600, 270)
    local explosive = explosivesFacade.createNewExplosive(request.explosiveId, request.presetId, vec3(request.position))
    explosivesHandler.addExplosive(request.uiId, request.explosiveId, request.presetId, explosive)
  end)
end

--- Cleans up all handlers, and sets the `uiNeedsResetting` flag to `true`
local function onFreeroamLoaded()
  worldObjectsHandler.initialise()
  vehiclesHandler.clear()
  explosivesHandler.clear()
  previewsHandler.clear()
  collisions.onMapLoaded()
  uiNeedsResetting = true
end

--- @return { explosives: ExplosiveIdentifier[], bossData: BossData, isPlayerInAVehicle: boolean, needsReset: boolean }
local function getCurrentState()
  local doesUiNeedResetting = uiNeedsResetting
  uiNeedsResetting = false -- Reset to false so this is only returned once for the UI

  return {
    explosives = explosivesHandler.getIds():toPrimitive(),
    bossData = explosivesHandler.getBossData(),
    isPlayerInAVehicle = vehiclesHandler.isPlayerInAVehicle(),
    needsReset = doesUiNeedResetting
  }
end

return {
  -- BeamNG interface
  onPreRender = freeroamHookWrapper(onPreRender),
  onVehicleSpawned = freeroamHookWrapper(vehiclesHandler.onVehicleSpawned),
  onVehicleResetted = freeroamHookWrapper(vehiclesHandler.onVehicleResetted),
  onVehicleDestroyed = freeroamHookWrapper(vehiclesHandler.onVehicleDestroyed),
  onVehicleSwitched = freeroamHookWrapper(vehiclesHandler.onVehicleSwitched),
  onFreeroamLoaded = onFreeroamLoaded,

  -- UI Interface
  getCurrentState = getCurrentState,
  placeExplosives = placeExplosives,
  getListOfExplosives = explosivesFacade.getListOfExplosives,
  createPreviewByOption = previewsHandler.createPreviewByOption,
  createPreviewByPosition = previewsHandler.createPreviewByPosition,
  updatePreviewPosition = previewsHandler.updatePreviewPosition,
  updatePreviewPositionByType = previewsHandler.updatePreviewPositionByType,
  deletePreview = previewsHandler.deletePreview,
}
